header { 
	package dspl.frontend;
	
	import dspl.intermediate.*;
	import dspl.intermediate.instructions.*;
	import dspl.intermediate.variable.*;
	import dspl.SymbolTable;
	import antlr.TreeParser;
	import antlr.collections.AST;
	import java.util.*;
}

/**
 * This class is the tree walker which walks the AST for a dspl file and produces an
 * intermediate representation.
 * @author Varun, Jeff, Mike, David
 */
class DSPLAstToIntermediateWalker extends TreeParser;
options {
	importVocab=ParserTokens; //we need to tkae the vocab from the parser and use that when walking the tree
}
{
	InstructionSet iSet = new InstructionSet(); //This InstructionSet is the root set for the whole program
	SymbolTable table = new SymbolTable(); //This is the symbol table for our only scope
	int nextTempNumber = 0; //This is a counter to keep track of what number our next temporary variable should be
	int rightType = -1;
	boolean convert = false;
	int indexCount = 0;
	public SymbolTable getSymbolTable() { 
		return table; 	
	}
	
	
	/**
	 * Helper function intended to be used by expressions which require a temporary variable to store a result
	 * variableType must correspond to one of the Variable.DATATYPE_xxx (int/uint/float/etc)<br>
	 * 
	 * @author dlariviere, jcropsey
	 */
	protected Variable getTempAtomicVar(int variableType, String prepend) throws SemanticAnalysisException { 
		
		Variable tempValue;
		try
		{
		tempValue = Variable.getNewAtomic(Variable.getAntlrTypeFromVarType(variableType), prepend+"_temp"+(nextTempNumber++), 0, false);
		try{
		table.addVariable(tempValue);
				
		}
		catch(SymbolicException e)
		{
			tempValue = getTempAtomicVar(variableType, prepend+"_");
		}
		}
		catch(NumberFormatException e)
		{
			throw new SemanticAnalysisException(null, "Number too large for assignment");
		} 
		
		return tempValue;
	}
	
	/**
	* Series of helper functions for conversion analysis
	* @author JCropsey
	*/
	protected void incrementIndexCount()
	{
		indexCount++;
	}
	
	protected void decrementIndexCount()
	{
		indexCount--;	
	}
	
	protected boolean isIndex()
	{
		if(indexCount > 0)
			return true;
		return false;
	}
	
	protected void setConversion(int conversionType)
	{
		rightType =	conversionType;
		convert = true;
	}
	
	protected void disableConversion()
	{
		convert = false;
	}
	
	protected int findType(int dataType)
	{
		if(isIndex())
			return Variable.DATA_TYPE_INT;
		else if(convert)
			return rightType;
		else
			return dataType;
	}
	
	protected boolean needsConversion(int dataType)
	{	
		if((convert && dataType != rightType && !isIndex())) { //dlariviere - added !isIndex, also your converting based on past datatype and not 
			return true;
		}
		if (isIndex() && dataType != Variable.DATA_TYPE_INT) { 
			return true; 	
		}
		return false;	
	}
	
	/**
	*Function that creates a new Float "Constant" (since its stored in the symbol table)
	*@author JCropsey
	*/
	protected Variable newFloatConstant(String val) throws SemanticAnalysisException
	{
		try
		{
			Variable myReal = Variable.getNewAtomic(Variable.getAntlrTypeFromVarType(Variable.DATA_TYPE_FLOAT),table.getNextFloatConstID(), 1, false); 
			System.out.println("Created Real Var");
			myReal.setValue(val); //set the value of the variable to the floating poitn constant string. doing so should get it printed in .data by the AsmFileGen
			System.out.println("Set Variable Value to: " + myReal.getValue());
			try
			{
	 	 		table.addVariable(myReal);  // @author Dave, actually add the variable to the symbol table
	  			System.out.println("Adding to Real to the symbol table");
	  		} 
		  	catch(SymbolicException e)
		  	{ 
		  		throw new Error("Failed on adding floating point Constant to ST: " + e.getMessage());
	  		}
	  	
	  	//added dlariviere to ensure conversion
		  	if (needsConversion(myReal.getDataType())) 
		  	{ 
				Variable tempConvert; 
				tempConvert = getTempAtomicVar(findType(myReal.getDataType()), ""); //modified dlariviere - needed findType(), else it wasn't actually converting to the right type, but instead to itself
				iSet.append(new ArithmaticInstruction(ArithmaticInstruction.OPERATION_CONVERSION, myReal, null, tempConvert));
				myReal = tempConvert;
	  		}
	  		return myReal;
		}
		catch(NumberFormatException e)
		{
			throw new SemanticAnalysisException(null, "Number too large for assignment");
		}
	}
	
	/**
	*Functions that Assist in Semantic Anaylsis or Array operations
	*@author JCropsey
	*/
	
	protected int arraysize = 0;
	protected boolean arrayop = false;
	
	protected void setSize(int size)
	{
		arraysize = size;	
	}
	
	protected int getSize()
	{
		return arraysize;	
	}
	
	protected boolean validArrayOperation(int size)
	{
		if(size == arraysize)
			return true;
		return false;
	}
	
	protected void arrayOperationEnter()
	{
		arrayop = true;	
	}
	
	protected void arrayOperationExit()
	{
		arrayop = false;	
	}
	
	protected boolean isArrayOp()
	{
		return arrayop;	
	}
	
	/**
	*Functions to aid in Semantic checking of logical expressions
	*@author JCropsey
	*/
	protected int checkLogical(Variable e1, Variable e2, AST _t) throws SemanticAnalysisException
	{
		if(!(!e1.isConstant() && !e2.isConstant()) && !(e1.getDataType() == e2.getDataType()))
			if(e1.isConstant())
				return 1;
			else
				return 2;
		if(e1.getDataType() != e2.getDataType())
			throw new SemanticAnalysisException((DSPLAST) _t, "Arguments to a logical operation must be of the same type");
		if(e1.isArray() || e2.isArray())
			throw new SemanticAnalysisException((DSPLAST) _t, "Arguments to a logical operation	cannot be arrays");
		return 0;
		
	}
	
}

/**
 * The program rule is the one that...matches the whole program, represented by the root node
 * of the AST.
 * The program rule is the one that dsplc calls and so it must return the root instruction
 * set.
 *
 * When writing rules it is important to understand how they work.  ANTLR, when tree-walking
 * uses a syntax of:
 * variable=rule {java action}
 * only the rule is required, but it is possible, depending on what the rule return, to set
 * its return value to a variable, and also set some action that will be performed when the
 * the rule has been matched.  This action is performed *after* the rule is matched.
 */
program returns [InstructionSet set] throws SemanticAnalysisException, SymbolicException //set is the object being returned
{
	InstructionSet pset = null;
	pset = null;

	Instruction s;
	set = iSet; 
	try
	{
		table.constructTable(_t); //build the symbol table from the AST
	}
	catch(NumberFormatException e)
	{
		throw new SemanticAnalysisException(null, "Number too large for assignment");
	}
}
: 
#(PROGRAM		//PROGRAM is the root, its children are the decls and a series of statments
(decls)?			//this then calls the 'decls' rule
(s=stmts)?			//this calls the 'stmts' rule
)
;

/**
 * We construct the symbol table outside of the tree walker, so we don't actually do anything
 * for the declarations that we find here
 */
decls
{
}
: #(DECLS
(ID)*)
;

/** stmts is simply a sequence of statments */
stmts returns [Instruction s] throws SemanticAnalysisException, SymbolicException
{ s = null; Instruction s1;}
: s=stmt (s1=stmts)?
;

/** a statements can be one of many types, and an appropriate instruction needs to be
 * appended to the current InstructionSet based on what needs to be done
 */
stmt returns [Instruction s] throws SemanticAnalysisException, SymbolicException
{ Variable e1, e2;
  s = null;
	Instruction s1, s2;
	Comparison sb;
	
}
: #(ASSIGN e1=expr {setConversion(e1.getDataType()); 
	if(e1.isArray()){arrayOperationEnter(); setSize(e1.getArraySize());}}
	e2=expr {
		disableConversion(); arrayOperationExit();
		}//An assignment takes two expressions, one that we are evaluating, the other we are assigning the result to
	{ 	System.out.println("ASSIGN Found");
		System.out.println("Going to add instruction.");
		//checking for pattern where temporary variable is created to store the results of an expression before assigning it to another variable
		//if pattern is matched, the expression is condensed into a single expression, with the actual destination being plugged in place of the temp
		//the only exception to this is when using elements whihc require temporary registers 
		if ((iSet.last() instanceof ArithmaticInstruction) &&  //if the last instruction added was an ArithmaticInst
			e2 == ((ArithmaticInstruction) iSet.last()).dest 	//and if the last instruction added was writing a value to e2
			&& !e1.isElement()) {								//and e1 is not an element Variable 
			
			((ArithmaticInstruction) iSet.last()).dest = e1;	//then instead of creating a new instruction, write the value of the last instruction to e1, instead of temp var
		}
		//else create a new instruction 
		else {
			s = new ArithmaticInstruction(ArithmaticInstruction.OPERATION_ASSIGNMENT,
				e2,null,e1);
			iSet.append(s);			
		}
	}
) 
|
  //For ccall, we first match the 'ccall' literal, then we fetch the function name, and then the list of arguments, one by one
  #(LITERAL_ccall { System.out.println("Going to add ccall"); String functionName = null; Variable currentArg = null; }
  	functionName=ccallname { s = new CCallInstruction(functionName); }
  	(currentArg=expr {((CCallInstruction) s).addParameter(currentArg);})*
  	{ iSet.append(s); }
)
| 
  #(LITERAL_if { System.out.println("Going to add if"); s2 = null; } 
	 sb=bool s1=subprog (s2=subprog)? 
	 	{
	 		Comparison sr = ((Comparison)sb);
	 		s = new BranchInstruction(new Comparison(sr.arg1, sr.arg2, sr.condition), s1, s2);
	 		iSet.append(s);
	 	}
)
|
  #(LITERAL_for {System.out.println("Going to add for"); Instruction incrementInstruction;}
  	(stmt)* {InstructionSet superIset = iSet; iSet = new InstructionSet();} sb=bool {s1 = iSet.first(); iSet = new InstructionSet();}
  	incrementInstruction=stmts {s1 = iSet.first(); iSet = superIset;} s2=subprog
  	{	s = new LoopInstruction(null, new Comparison(sb.arg1, sb.arg2, sb.condition), s1, s2);
  		iSet.append(s);}
)
| 
  #(LITERAL_while { System.out.println("Going to add while");} 
	 sb=bool s1=subprog
	 {
	 	Comparison sr = ((Comparison)sb);
	 	s = new LoopInstruction(null, new Comparison(sr.arg1, sr.arg2, sr.condition), null, s1);
	 	iSet.append(s);
	 }
)
;

/**ccallname is unlike other instances of ID because we don't want a variable, we just want
 * the name of the function!
 */
ccallname returns [String name]
{
	name = null;
}
: ID {name = #ID.getText();}
;

subprog returns [Instruction s] throws SemanticAnalysisException, SymbolicException
{
	InstructionSet superIset = iSet;
	iSet = new InstructionSet();
	s = null;
}
:
#(LBRACE (s=stmts)) {s = iSet.first(); iSet = superIset;}
;
//@author jcropsey
bool returns [Comparison s] throws SemanticAnalysisException, SymbolicException
{
	Variable e1, e2;
	s = null;
}
:
#(GT e1=expr e2=expr	//Greater Than takes two expressions
	{
		Variable exchange = null;
		System.out.println("Going to add greater than");
		switch(checkLogical(e1,e2,_t))
		{
			case 1: exchange = getTempAtomicVar(e2.getDataType(), "");
					exchange.setValue(String.valueOf(e1.getValue()));
					e1 = exchange;
					break;
			case 2: exchange = getTempAtomicVar(e1.getDataType(), "");
					exchange.setValue(String.valueOf(e2.getValue()));
					e2 = exchange;
					break;
			default: //nothing
					break;
		}
		s = new Comparison(e1, e2, Comparison.CONDITION_GREATER);
	}
)
|
#(LT e1=expr e2=expr	//Less Than takes two expressions
	{
		Variable exchange = null;
		System.out.println("Going to add less than");
		switch(checkLogical(e1,e2,_t))
		{
			case 1: exchange = getTempAtomicVar(e2.getDataType(), "");
					exchange.setValue(String.valueOf(e1.getValue()));
					e1 = exchange;
					break;
			case 2: exchange = getTempAtomicVar(e1.getDataType(), "");
					exchange.setValue(String.valueOf(e2.getValue()));
					e2 = exchange;
					break;
			default: //nothing
					break;
		}
		s = new Comparison(e1, e2, Comparison.CONDITION_LESS);
	}
)
|
#(GE e1=expr e2=expr	//Greater Than or Equal To takes two expressions
	{
		Variable exchange = null;
		System.out.println("Going to add greater than or equal to");
		switch(checkLogical(e1,e2,_t))
		{
			case 1: exchange = getTempAtomicVar(e2.getDataType(), "");
					exchange.setValue(String.valueOf(e1.getValue()));
					e1 = exchange;
					break;
			case 2: exchange = getTempAtomicVar(e1.getDataType(), "");
					exchange.setValue(String.valueOf(e2.getValue()));
					e2 = exchange;
					break;
			default: //nothing
					break;
		}
		s = new Comparison(e1, e2, Comparison.CONDITION_GE);
	}
)
|
#(LE e1=expr e2=expr	//Less Than or Equal To takes two expressions
	{
		Variable exchange = null;
		System.out.println("Going to add less than or equal to");
		switch(checkLogical(e1,e2,_t))
		{
			case 1: exchange = getTempAtomicVar(e2.getDataType(), "");
					exchange.setValue(String.valueOf(e1.getValue()));
					e1 = exchange;
					break;
			case 2: exchange = getTempAtomicVar(e1.getDataType(), "");
					exchange.setValue(String.valueOf(e2.getValue()));
					e2 = exchange;
					break;
			default: //nothing
					break;
		}
		s = new Comparison(e1, e2, Comparison.CONDITION_LE);
	}
)
|
#(EQ e1=expr e2=expr	//Equal To takes two expressions
	{
		Variable exchange = null;
		System.out.println("Going to add equal to");
		switch(checkLogical(e1,e2,_t))
		{
			case 1: exchange = getTempAtomicVar(e2.getDataType(), "");
					exchange.setValue(String.valueOf(e1.getValue()));
					e1 = exchange;
					break;
			case 2: exchange = getTempAtomicVar(e1.getDataType(), "");
					exchange.setValue(String.valueOf(e2.getValue()));
					e2 = exchange;
					break;
			default: //nothing
					break;
		}
		s = new Comparison(e1, e2, Comparison.CONDITION_EQUAL);
	}
)
|
#(NE e1=expr e2=expr	//Greater Than or Equal To takes two expressions
	{
		Variable exchange = null;
		System.out.println("Going to add not equal to");
		switch(checkLogical(e1,e2,_t))
		{
			case 1: exchange = getTempAtomicVar(e2.getDataType(), "");
					exchange.setValue(String.valueOf(e1.getValue()));
					e1 = exchange;
					break;
			case 2: exchange = getTempAtomicVar(e1.getDataType(), "");
					exchange.setValue(String.valueOf(e2.getValue()));
					e2 = exchange;
					break;
			default: //nothing
					break;
		}
		s = new Comparison(e1, e2, Comparison.CONDITION_NEQUAL);
	}
)
;

expr returns [Variable object] throws SemanticAnalysisException, SymbolicException
{
	object = null;
	//For expressions that require some match, we must add the instruction to be performed
	//to the current instruction set.  Then we can just return the Variable the represents the
	//result.
	Instruction exprInstruct;
	System.out.println("Matching expr");
	
}
:
//JCropsey 12-15-06 - added negation (with help of VKM)
#(NEGATE {Variable negative; Variable myVar = null;} negative=expr
{	
	if(negative.getDataType() == Variable.DATA_TYPE_FLOAT)
	{
		System.out.println("Making Negative Real"); 
		myVar = newFloatConstant("-"+String.valueOf(negative.getValue()));
	}
	else if(negative.getDataType() != Variable.DATA_TYPE_FLOAT && negative.isConstant())
	{
		try
		{
			myVar = Variable.getNewAtomic(Variable.getAntlrTypeFromVarType(findType(0)), "-" + negative.getIdentifier(), 1, true);
		}
		catch(NumberFormatException e)
		{
			throw new SemanticAnalysisException(null, "Number too large for assignment");
		}
	}
	else
	{
		throw new SemanticAnalysisException((DSPLAST)_t, "Only constants can be negated");	
	}
	object = myVar;
} )
|
#(PLUS
	{
		Variable arg1, arg2;
		System.out.println("Going to match PLUS args");
	}
	arg1=expr arg2=expr //we match the two arguments for PLUS
	{
		object = getTempAtomicVar(findType(0), ""); //@author dlariviere - utilizes helper function instead of repeating the same code over and over
		exprInstruct = new ArithmaticInstruction(ArithmaticInstruction.OPERATION_ADDITION, arg1, arg2, object);
		iSet.append(exprInstruct);
	}
) |
#(MINUS
	{
		Variable arg1, arg2;
		System.out.println("Going to match PLUS args");
	}
	arg1=expr arg2=expr //we match the two arguments for MINUS
	{
		object = getTempAtomicVar(findType(0), ""); //@author dlariviere - utilizes helper function instead of repeating the same code over and over
		exprInstruct = new ArithmaticInstruction(ArithmaticInstruction.OPERATION_SUBTRACTION, arg1, arg2, object);
		iSet.append(exprInstruct);
	}
) |
#(MUL
	{
		Variable arg1, arg2;
		System.out.println("Going to match MUL args");
	}
	arg1=expr arg2=expr //we match the two arguments for Multiplication
	{
		object = getTempAtomicVar(findType(0), ""); //This is the temporary variable we will use to store the result.
		exprInstruct = new ArithmaticInstruction(ArithmaticInstruction.OPERATION_MULTIPLICATION, arg1, arg2, object);
		iSet.append(exprInstruct);
	}
) 
|
#(DIV
	{
		Variable arg1, arg2;
		System.out.println("Going to match DIV args");
	}
	arg1=expr arg2=expr //we match the two arguments for Division
	{
		object = getTempAtomicVar(findType(0), ""); //This is the temporary variable we will use to store the result.
		exprInstruct = new ArithmaticInstruction(ArithmaticInstruction.OPERATION_DIVISION, arg1, arg2, object);
		iSet.append(exprInstruct);
	}
) 
|

//added by Dave - 2006-12-8: Handles ID[index] where index is a constant (num) or another ID
//modified by Jeffrey = 2006-12-13: Attempting conversion
#(ID {
		int numChildren = #ID.getNumberOfChildren(); //number of children of this indicates whether it is a simple id, or an id[expr], where expr is a child node
		Variable indexVar = null; 		//only used if #ID is followed by exactly one expression as its child 
		Variable parentVar = null; 		//only used if indexed ID, contains the parent
		Variable tempConvert = null; 	//used only if conversion is necessary
		Variable tempCheck = null; 		//holds var object during analysis
		
		//if no children, just an identifier (no [0] or .real, etc) 
		if (numChildren==0) { 
			tempCheck = table.getVariable(#ID.getText());
			object = tempCheck;
			if(isArrayOp() && (!(tempCheck.isArray() && tempCheck.getDataType() == findType(0) && tempCheck.getArraySize() == getSize()) || !tempCheck.isArray()))
			{
				throw new SemanticAnalysisException(null, "Arrays in array operations must be of the same length; array/atomic arithmatic operations not supported");
			}
			else if(needsConversion(tempCheck.getDataType()))
			{
				System.out.println(#ID.getText() + " needs a conversion from " + tempCheck.getDataType() + " to " + findType(tempCheck.getDataType())); 
				if (findType(tempCheck.getDataType())==tempCheck.getDataType()) { 
					throw new Error("converting to same data type"); 	
				}
				tempConvert = getTempAtomicVar(findType(tempCheck.getDataType()), "");
				iSet.append(new ArithmaticInstruction(ArithmaticInstruction.OPERATION_CONVERSION, tempCheck, null, tempConvert));
				object = tempConvert;
			}			
		}
		else if (numChildren>3) { //greater than 3 since there is LBRACK (actual expression) RBRACK
			throw new Error("Error: more than one expression following an identifier, must be zero or one (\"[0]\" or \".real\" for example)"); 	
		}
		else { //one child, must be either an index into the element, or a .real/.imag first complex
			//tempValue = Variable.getNewAtomic(Variable.LITERAL_int, "_temp"+(nextTempNumber++), 0, false);
			parentVar = table.getVariable(#ID.getText());
			
			//table.addVariable(tempValue);
			//object = tempValue; //This is the temporary variable we will use to store the result.
		}
	} 
	//handles case where ID is followed by an expression, which must be the index into the expression (and expression must be indexable, either with [0] or .real/.imag)
	//TODO: (TreeWalker People) the way things are now, an ID followed by ANY EXPRESSION (not just [0]/.real/.imag will get parsed. We need way to differentiate, should probably be added to the grammar as well
	(LBRACK {incrementIndexCount();} indexVar=expr RBRACK {decrementIndexCount();}
		{ 
			if (indexVar.isConstant() && indexVar.getIdentifier().substring(0,1).compareTo("-") == 0)
				throw new SymbolicException(parentVar, " : Array index accesses cannot be negative");
			if (indexVar.isArray())
				throw new SymbolicException(indexVar, " : An array cannot be the index of an array");
			if (indexVar.isElement()) { //handle if expression is another indexed id, for example "a[b[0]] = 5;" 
				//move the element ("b[0]", in the example above) into a temporary variable and then use that variable as the new index element
				Variable tempIndexValue = getTempAtomicVar(Variable.LITERAL_int, "");
				exprInstruct = new ArithmaticInstruction(ArithmaticInstruction.OPERATION_ASSIGNMENT, indexVar, null, tempIndexValue);
				iSet.append(exprInstruct);
				indexVar = tempIndexValue; //the index into the expression is now the temp variable
				//throw new Error("Error: indexes must be either constants or atomic elements, double indexing (\"a[b[i]] = 5;\") is not yet implemented\n"); 	
			}
			try { 
				Variable elementVar = Variable.getNewAtomicElement(parentVar, indexVar, 0); 
				object = elementVar;
				if(needsConversion(elementVar.getDataType()))	//Checks to see if conversion is necessary based on save type
				{
					System.out.println("Converting array element var from " + elementVar.getDataType() + " to " + findType(elementVar.getDataType())); 
					tempConvert = getTempAtomicVar(findType(elementVar.getDataType()), ""); //modified dlariviere - needed findType(), else it wasn't actually converting to the right type, but instead to itself
					iSet.append(new ArithmaticInstruction(ArithmaticInstruction.OPERATION_CONVERSION, elementVar, null, tempConvert));
					object = tempConvert;
				}
				
			} catch (Exception ex) { //catch a possible exception thrown
				throw new SemanticAnalysisException(null,"Array index out of bounds at ");  //throw the only type of expression that ANTLR's walker is already throwing 
			}
			
		}
	)?
)
|
(NUM // If we see a constant number we create a new Variable 
	{ 
		try
		{
			if(findType(1) != Variable.DATA_TYPE_FLOAT)
			{
				object = Variable.getNewAtomic(Variable.getAntlrTypeFromVarType(findType(1)), #NUM.getText(), 1, true); 
			}
			else
			{
				object = newFloatConstant(#NUM.getText());
			}
		}
		catch(NumberFormatException e)
		{
			throw new SemanticAnalysisException(null, "Number to large for assignment");
		}
	}
) |
(REAL {Variable myReal = null;} // If we see a constant floating point number we create a new Variable of constant type (note in this context real is only float. all natural num's should be handled by NUM above)
	{ 
		/**
		 * Added by dlariviere. x87 FPU has no load immediate for arbitrary constnats (only a few special cases)<br>
		 * Therefore float constants, unlike all other constants, must be added to the sybmol table and treated as actual variables, not constants. 
		 */
		System.out.println("Found real: " + #REAL.getText());
		object = newFloatConstant(#REAL.getText());
	}
) 
|
(STRING //When constructing the symbol table before, we did not add strings, so we'll do that now
	{
	  	try
	  	{
	  		object = table.addStringVariable(#STRING.getText());  // @author Dave, actually add the variable to the symbol table
	  	}
	  	catch(SymbolicException e)
	  	{ 
	  		throw new SemanticAnalysisException((DSPLAST)_t,"Failed on String Constant: " + e.getMessage());
	  	}
	}
) 
;