// Here is the grammar for DSPL, both Parser and Lexer
/*
 * Add the header below to the top of every generated file; specifically, make them all part of the dspl.frontend package
 */
header { package dspl.frontend; import antlr.*;}

class DSPLParser extends Parser;
options {
	buildAST = true;
	k = 2;
	exportVocab=ParserTokens;
}
tokens { NEGATE; DECLS; PROGRAM; }
{
	public void reportError(antlr.RecognitionException e) {
		System.err.println(e.toString());
		System.exit(2);
	}
}

/* A program must start first with declarations, then a series of statements */
program :	(decls)? (stmt)* { #program = #([PROGRAM, "PROGRAM"], #program);};
/* subprograms are for example the bodies of if() statements and for() loops */
subprogram : LBRACE^ (stmt)* RBRACE!;
/* "DECLS" is the root with children consisting of each declaration being made in the program */
decls :    (decl)* {
				#decls = #([DECLS, "DECLS"], #decls);
				};

/** Variable declarations are allowed with names like:
	char foo;
	complex d;
	complex_array[10] fo_o;
	int_array[3] someArray3;
*/
decl :  ( ("string" | "int" | "uint" | "byte" | "ubyte" | "float" | "complex") ID^ SEMI!
		| ("complex_array" | "int_array" | "uint_array" | "byte_array" | "ubyte_array" | "float_array") (LBRACK! NUM RBRACK!) ID^ SEMI!
		) ;

stmt : 	    assignment SEMI!
			| "if"^ LPAREN! bool RPAREN! subprogram
				(options {greedy=true;}: "else"! subprogram )?
			| "while"^ LPAREN! bool RPAREN! subprogram
			| "for"^ LPAREN! (assignment (COMMA! assignment)*)? SEMI! bool SEMI! (assignment (COMMA! assignment)*)? RPAREN! subprogram
			| ccall
			| SEMI!
			;
ccall :		"ccall"^ ID LPAREN! (unary (COMMA! unary)*)? RPAREN! SEMI!;
assignment: loc ASSIGN^ expr;

/*	The block that follows is our core algebra of the grammar,
	it manages important things like operator precedence, associativity
	and array accesses */
bool : 	   	unary ((EQ^ | NE^ | LT^ | LE^ | GT^ | GE^) unary)? ;
expr : 	   	term ((PLUS^ | MINUS^) term)* ;
term : 	   	paren ((MUL^ | DIV^) paren)* ;
paren:		(LPAREN! expr RPAREN!) | unary;
unary :    	MINUS^ (NUM | REAL) { #unary.setType(NEGATE);}
			| PLUS! unary | factor ;
factor :   	loc | NUM | REAL | STRING ;
loc :	   	ID^ (LBRACK expr RBRACK)?; //2006-12-09 @author dave, LBRACK/RBRACK are kept in the tree, required in tree parser to recognize indexing versus any other expression
/* End core algebra */

class DSPLLexer extends Lexer;
options {
	k = 2;
	charVocabulary = '\3'..'\377';
	testLiterals = false;
} 

protected
NEWLINE
	: ( '\r' (options {greedy=true;}:'\n')?  // DOS & Windows & Macintosh
      | '\n'    // Unix
      )
    // increment the line count in the scanner, keeps error line counters in sync
    { newline(); }
	;

protected
DIGITS
	: ('0'..'9')+ ;
NUM 
	: DIGITS ('.' DIGITS {_ttype = REAL;})? ;

/*	Identifiers must start with either a letter or underscore and then can be of
	arbitrary length of letters, undescores, and digits */
ID options {testLiterals = true;}
	: ('a'..'z' | 'A'..'Z' | '_') ('a'..'z' | 'A'..'Z' | '_' | '0'..'9')*;

LE : "<=" ; SEMI : ';' ; 
GT : '>' ; LPAREN : '(' ; 
ASSIGN : '=' ; GE : ">=" ; RPAREN : ')' ; 
EQ : "==" ; LBRACE : '{' ; PLUS : '+' ; 
RBRACE : '}' ; MINUS : '-' ; 
NE : "!=" ; LBRACK : '[' ; MUL : '*'; 
LT : '<' ; RBRACK : ']' ; DIV : '/';
COMMA : ','; TILDE : '~'; DOT: '.';

// TODO: Add support for escaping characters in strings
STRING
  :	'"'
  	( // Prevent .* from eating the whole file 
			options {greedy=false;}: 
			( 	NEWLINE
				|  ~( '\n' | '\r' | '"')
			) 
		)*
	'"' ;

WS
  : ( ' '
    | '\t'
    | '\f'
    // handle newlines
    | NEWLINE
    )
    { $setType(Token.SKIP); }
  ;

COMMENT
	: '/' 
		( ('*') => '*'
			( // Prevent .* from eating the whole file 
				options {greedy=false;}: 
				( 	NEWLINE
					|  ~( '\n' | '\r' )
				) 
			)* 
	   	  "*/"
	   	  | '/' ( ~('\n' | '\r'))* (NEWLINE)?
		) 
	{ $setType(Token.SKIP); }
    ;  